// line numbers are occasionally way off
// compile option to speed things up
// errors in a SWITCH aren't found in the test
// "if(x) cmd" behaves strangely if {} are omitted around cmd

#include "Common/StdAfx.h"
#ifndef _CONSOLE
#include "Kernel.h"
#define doprint(x)
#else
#define doprint(x) printf(x) // this is sloppy
#endif
#include "Compress.h"
#include "Parse.h"
#include "../Strings.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#ifndef NO_PARSER

CParser Script;

CParser::CParser (void) {
    memset (this, 0, sizeof (CParser));
}

CParser::~CParser (void) {
    CVar *temp = varlist;
    while (temp) {
        temp = temp->next;
        if (varlist->arraydata) delete [] varlist->arraydata;
        delete varlist;
        varlist = temp;
    }
    if (script && free_scr) delete [] script;
}

bool CParser::ReadScript (unsigned offset, unsigned scrnum, char *ptr) {
    unsigned i, j;
    FILE     *file;
    CVar     *temp = varlist;
    char     buf[32];
    while (temp) {
        temp = temp->next;
        if (varlist->arraydata) delete [] varlist->arraydata;
        delete varlist;
        varlist = temp;
    }
    if (script && free_scr) delete [] script;
    i = tracelevel;
    j = number;
    memset (this, 0, sizeof (CParser));
    tracelevel = i;
    number     = j;
    for (i=Str_W_Last + 1; i<Str_SysLast; i++) if (StrTable[i]) {
        AddVar (&(StrTable[i][1]), P_NewFunc); // AddVar (i, P_NewFunc);
        ((CVar*) varptr)->arraysize = i;
        ((CVar*) varptr)->value     = StrTable[i][0] - '0';
    }
    if (ptr) {
        free_scr = false;
        script   = ptr;
        for (; script[scrlen]; scrlen++);
    } else if (theApp.MediaLib == NULL) {
        // another large debug-only routine
        free_scr = true;
        theApp.ChangeDirectory (theApp.HomeDir);
        strcpy (buf, StrTable[Str_BaseHelp]);
        for (i=0; buf[i]; i++);
        for (i--; buf[i] != '\\'; i--);
        buf[i] = 0;
        theApp.ChangeDirectory (buf);
        file = fopen (StrTable[Str_Script + scrnum], "rb");
        if (file == NULL) return false;
        fseek (file, 0, SEEK_END);
        scrlen = ftell (file);
        rewind (file);
        script = new char [scrlen];
        if (script == NULL) {
            theApp.runlevel = Run_MemError;
            return false;
        }
        fread (script, scrlen, 1, file); 
        if (ferror (file)) return false;
        fclose (file);
    } else {
        free_scr = true;
        i = MediaRef[offset + Str_Script - Str_MiscFile + scrnum];
        if (i == 0) {
            theApp.runlevel = Run_FileError;
            return false;
        }
        fseek (theApp.MediaLib, i, SEEK_SET);
        script = pack.ExpandToRam (theApp.MediaLib);
        scrlen = pack.textsize;
    }
    // MakeByteCode ();

    while (1) {
        i = Process ();
        if (i == Str_S_None) {
            if (mainpos) {
                skip   = 0;
                end    = 0;
                pos    = mainpos;
                action = Str_W_Main;
                PushScope (P_Command);
                break;
            }
            i = Str_S_NoMain;
        }
        if (i) {
            PrintError (i);
            return false;
        }
    }
    sprintf (outbuf, StrTable[Str_S_None]);
    return true;
}

void CParser::Output (const char *text) {
    unsigned i, j;
    doprint (text);
    for (i=0; text[i]; i++) if (text[i] == 0x0a) outline ++;
    i=0;
    if (outline > 9) {
        while (outline > 9) {
            if (outbuf[i] == 0x0a) outline --;
            if (outbuf[i] == 0)    break;
            i++;
        }
        for (j=0; outbuf[i]; i++, j++) outbuf[j] = outbuf[i];
    } else {
        for (j=0; outbuf[j]; j++);
    }
    for (i=0; text[i]; i++, j++) outbuf[j] = text[i];
    outbuf[j] = 0;
}

void CParser::Trace (unsigned lev, const char *str, const char *e1, const char *e2) {
    if (tracelevel >= lev && str) {
        TRACE1 ("[%3.3d] ", curline);
        if      (e2) TRACE2 (str, e1, e2);
        else if (e1) TRACE1 (str, e1);
        else         TRACE0 (str);
    }
}

void CParser::PrintError (unsigned errnum) {
    char     b1  [1024];
    unsigned i, j=1, k=0;
    b1[0] = 0;
    if (script != NULL) for (i=0; script[i]; i++) {
        if (script[i] == 0x0d || script[i] == 0x0a) {
            j++;
            k = 0;
            if (i >= pos) break; // if (j > curline) break;
        } else if (script[i] >= ' ') {
            b1[k] = script[i];
            if (k < 1023) k++;
            b1[k] = 0;
        }
    }
    sprintf (outbuf, StrTable[Str_ErrParse], j - 1, b1, StrTable[errnum]);
    curline = j - 1; // no
    end     = true;
    error   = errnum;
}

void CParser::Start (void) {
    end        = 0;
    error      = 0;
    pos        = mainpos;
    scopelevel = 0;
    action     = Str_W_Main;
    outbuf[0]  = 0;
    Trace     (1, StrTable[Str_S_Main]);
    PushScope (P_Command);
}

unsigned CParser::FunctionCall (const char *funcname, unsigned p0, unsigned p1, unsigned p2) {
    CVar     *tmp;
    unsigned i, j;
    if (funcname == NULL) {
        Script.Start ();
        for (i=0; i<10000; i++) {
            j = Script.Process ();
            if (j == Str_S_None) break;
            else if (j != 0) {
                Script.PrintError (j);
                return false;
            }
        }
        return true;
    }
    for (tmp = varlist; tmp != NULL; tmp = tmp->next)
        if (strcmp (tmp->name, funcname) == 0) break;
    if (tmp == NULL) return 0;
    if      (p2) funcparmno = 3;
    else if (p1) funcparmno = 2;
    else if (p0) funcparmno = 1;
    else         funcparmno = 0;
    end         = 0;
    funcparm[0] = p0;
    funcparm[1] = p1;
    funcparm[2] = p2;
    retpos      = scrlen;
    action      = Str_W_Main;
    pos         = tmp->scope;
    value       = 0;
    PushScope (P_FuncEntry);
    for (i=0; i<10000; i++) {
        j = Process ();
        if (j == Str_S_None) break;
        else if (j != 0) {
            PrintError (j);
            break;
        }
    }
    if (i >= 10000) return 0;
    return value;
}

void CParser::SaveVars (void) {
    CVar     *temp;
    unsigned i;
    for (temp = varlist; temp != NULL; temp = temp->next) {
        if (temp->scope == 0 && temp->is_func == false) {
            if (temp->arraysize) {
                for (i=0; i<temp->arraysize; i++) pack.WriteD (temp->arraydata[i]);
            } else {
                pack.WriteW (temp->value);
            }
        }
    }
}

void CParser::LoadVars (void) {
    CVar     *temp;
    unsigned i;
    for (temp = varlist; temp != NULL; temp = temp->next) {
        if (temp->scope == 0 && temp->is_func == false) {
            if (temp->arraysize) {
                for (i=0; i<temp->arraysize; i++) temp->arraydata[i] = pack.ReadD ();
            } else {
                temp->value = pack.ReadW ();
            }
        }
    }
}

void CParser::AddVar (const char *word, unsigned type) {
    CVar *temp,
         *t2;
    int  i;
    if (word == NULL) return;
    temp = new CVar;
    if (temp == NULL) {
        theApp.runlevel = Run_MemError;
        return;
    }
    memset (temp, 0, sizeof (CVar));
    for (i=0; word[i]; i++) temp->name[i] = word[i];
    temp->name[i] = word[i];
    varptr        = &temp->value;
    switch (type) {
        case P_NewFunc:  temp->scope     = pos;
                         temp->is_func   = true;
                         break;
        case P_NewVar:   temp->scope     = scopelevel - 1;
                         temp->is_func   = false;
                         break;
    }
    if (varlist == NULL) {
        varlist = temp;
    } else {
        for (t2 = varlist; t2 != NULL; t2 = t2->next) {
            for (i=0; 1; i++) {
                if (word[i] < t2->name[i] || word[i] == 0) {
                    temp->next = t2;
                    if (t2 == varlist) {
                        varlist = temp;
                    } else {
                        for (t2 = varlist; t2->next != temp->next; t2 = t2->next);
                        t2->next = temp;
                    }
                    t2 = NULL;
                    break;
                }
                if (word[i] > t2->name[i]) break;
            }
            if (t2 == NULL) break;
            if (t2->next == NULL) {
                t2->next   = temp;
                temp->next = NULL;
                break;
            }
        }
    }
    Trace (2, StrTable[Str_S_AddVar], word, (char*) temp);
}

void CParser::AddVar (unsigned tref, unsigned type) {
    CVar *temp,
         *t2;
    if (tref == 0) return;
    temp = new CVar;
    if (temp == NULL) {
        theApp.runlevel = Run_MemError;
        return;
    }
    memset (temp, 0, sizeof (CVar));
    temp->tokenref = tref;
    varptr         = &temp->value;
    switch (type) {
        case P_NewFunc:  temp->scope     = pos;
                         temp->is_func   = true;
                         break;
        case P_NewVar:   temp->scope     = scopelevel - 1;
                         temp->is_func   = false;
                         break;
    }
    if (varlist == NULL) {
        varlist = temp;
    } else {
        for (t2 = varlist; t2 != NULL; t2 = t2->next) {
            if (tref <= t2->tokenref) {
                temp->next = t2;
                if (t2 == varlist) {
                    varlist = temp;
                } else {
                    for (t2 = varlist; t2->next != temp->next; t2 = t2->next);
                    t2->next = temp;
                }
                break;
            }
            if (t2->next == NULL) {
                t2->next   = temp;
                temp->next = NULL;
                break;
            }
        }
    }
    Trace (2, StrTable[Str_S_AddVar], (char*) tref, (char*) temp);
}

void CParser::KillVars (void) {
    CVar *temp,
         *prev;
    if (varlist) {
        while (varlist != NULL && varlist->scope > scopelevel && varlist->is_func == false) {
            temp = varlist;
            varlist = varlist->next;
            Trace (2, StrTable[Str_S_KillVar], temp->name, (char*) temp);
            if (temp->arraydata) delete [] temp->arraydata;
            delete temp;
        }
    }
    if (varlist) {
        prev = varlist;
        temp = prev->next;
        while (temp != NULL) {
            if (temp->scope > scopelevel && temp->is_func == false) {
                prev->next = temp->next;
                Trace (2, StrTable[Str_S_KillVar], temp->name, (char*) temp);
                if (temp->arraydata) delete [] temp->arraydata;
                delete temp;
                temp = prev->next;
            } else {
                prev = temp;
                temp = temp->next;
            }
        }
    }
}

void CParser::PushScope (unsigned type) {
    if (scopelevel >= MAX_SCOPE) {
        theApp.runlevel = Run_MemError;
    } else {
        memcpy (&scopestack[scopelevel], this, sizeof (CScope));
        scopelevel++;
        action = type;
        value  = 0;
        if (skip) skip = Skip_Stack;
    }
}

void CParser::PopScope (void) {
    char     buf[255];
    unsigned i, j;
    if (scopelevel == 0) {
        theApp.runlevel = Run_LogError;
    } else {
        scopelevel--;
        if (skip != Skip_Break) skip = scopestack[scopelevel].skip;
        retpos = scopestack[scopelevel].retpos;
        if (scopestack[scopelevel].action >= Str_W_Assign && scopestack[scopelevel].action <= Str_W_Dec) {
            if (scopestack[scopelevel].varptr == NULL) {
                theApp.runlevel = Run_LogError;
                return;
            }
            switch (scopestack[scopelevel].action) {
                case Str_W_Assign:                                                   break;
                case Str_W_AsgAdd:  value = *scopestack[scopelevel].varptr +  value; break;
                case Str_W_AsgSub:  value = *scopestack[scopelevel].varptr -  value; break;
                case Str_W_AsgMult: value = *scopestack[scopelevel].varptr *  value; break;
                case Str_W_AsgAnd:  value = *scopestack[scopelevel].varptr &  value; break;
                case Str_W_AsgOr:   value = *scopestack[scopelevel].varptr |  value; break;
                case Str_W_AsgXor:  value = *scopestack[scopelevel].varptr ^  value; break;
                case Str_W_AsgLSh:  value = *scopestack[scopelevel].varptr << value; break;
                case Str_W_AsgRSh:  value = *scopestack[scopelevel].varptr >> value; break;
                case Str_W_Inc:     value = *scopestack[scopelevel].varptr +      1; break;
                case Str_W_Dec:     value = *scopestack[scopelevel].varptr -      1; break;
                case Str_W_AsgDiv:  if (value == 0) value = 0xffffffff; else
                                    value = *scopestack[scopelevel].varptr /  value; break;
                case Str_W_AsgMod:  if (value != 0)
                                    value = *scopestack[scopelevel].varptr %  value; break;
            }
            if (value & 0x80000000) value = 0;
            if (value & 0x00010000) value = 0xffff;
            if (skip == 0) {
                if (tracelevel >= 1) {
                    sprintf (buf, StrTable[Str_S_SetVar], (char*) (scopestack[scopelevel].varptr) + 4, *scopestack[scopelevel].varptr, value);
                    Trace (1, buf);
                }
                *scopestack[scopelevel].varptr = value;
            }
            PopScope ();
            return;
        }
        switch (scopestack[scopelevel].action) {
            case Str_W_Power:   j=1;
                                for (i=0; i<value; i++) j *= scopestack[scopelevel].value;
                                value = j;                             action = P_ValueEnd; return;
            case Str_W_Divide:  if (value == 0) value = 0xffff; else
                                value = scopestack[scopelevel].value /  value; action = P_ValueEnd; return;
            case Str_W_Modulo:  if (value != 0)
                                value = scopestack[scopelevel].value %  value; action = P_ValueEnd; return;
            case Str_W_Mult:    value = scopestack[scopelevel].value *  value; action = P_ValueEnd; return;
            case Str_W_Add:     value = scopestack[scopelevel].value +  value; action = P_ValueEnd; return;
            case Str_W_Minus:   value = scopestack[scopelevel].value -  value; action = P_ValueEnd; return;
            case Str_W_And:     value = scopestack[scopelevel].value &  value; action = P_ValueEnd; return;
            case Str_W_Xor:     value = scopestack[scopelevel].value ^  value; action = P_ValueEnd; return;
            case Str_W_Or:      value = scopestack[scopelevel].value |  value; action = P_ValueEnd; return;
            case Str_W_LShift:  value = scopestack[scopelevel].value << value; action = P_ValueEnd; return;
            case Str_W_RShift:  value = scopestack[scopelevel].value >> value; action = P_ValueEnd; return;
            case Str_W_LogAnd:  value = scopestack[scopelevel].value && value; action = P_ValueEnd; return;
            case Str_W_LogOr:   value = scopestack[scopelevel].value || value; action = P_ValueEnd; return;
            case Str_W_Equals:  value = scopestack[scopelevel].value == value; action = P_ValueEnd; return;
            case Str_W_InEqual: value = scopestack[scopelevel].value != value; action = P_ValueEnd; return;
            case Str_W_GreatEq: value = scopestack[scopelevel].value >= value; action = P_ValueEnd; return;
            case Str_W_Greater: value = scopestack[scopelevel].value >  value; action = P_ValueEnd; return;
            case Str_W_LessEq:  value = scopestack[scopelevel].value <= value; action = P_ValueEnd; return;
            case Str_W_Less:    value = scopestack[scopelevel].value <  value; action = P_ValueEnd; return;
            case Str_W_Not:     value = !value;           action = P_ValueEnd; return;
            case Str_W_Negate:  value = (value ^ -1) + 1; action = P_ValueEnd; return;
            case P_CaseRun:
            case Str_W_BrackL:
                action = P_Command;
                scopelevel++;
                return;
            case Str_W_If:
                action = P_Then;
                if (skip == 0 && value == 0) skip = Skip_If;
                PushScope (P_Command);
                return;
            case Str_W_Switch:
                action = P_SwitchEnd;
                if (skip) {
                    PushScope (P_Switch);
                    action = P_SwBreak_0;
                } else {
                    PushScope (P_Switch);
                    action = P_Switch_0;
                }
                return;
            case P_SwitchEnd:
                if (skip == Skip_Break) skip = 0;
                PopScope ();
                return;
            case Str_W_Case:
                PopScope ();
                return;
            case P_Case:
                if (value == scopestack[scopelevel - 1].value) {
                    action = P_CaseRun;
                    PushScope (P_Command);
                } else {
                    action = P_Switch;
                    value = scopestack[scopelevel].value;
                }
                return;
            case Str_W_Else:
                PopScope ();
                return;
            case Str_W_While:
                action = P_Loop;
                if (skip == Skip_Break
                || (skip == 0 && value == 0)) skip = Skip_While;
                PushScope (P_Command);
                return;
            case P_For4:
                pos = scopestack[scopelevel].retpos;
                action = P_For0;
                PushScope (P_Command);
                skip = Skip_For1;
                return;
            case P_For0:
                action = P_For1;
                PushScope (P_Value);
                return;
            case P_For1:
                action = P_For2;
                if (skip == Skip_Break
                || (skip == 0 && value == 0)) skip = Skip_For;
                PushScope (P_CmdVar);
                if (skip == 0) skip = Skip_For1;
                return;
            case P_FuncRef1:
                funcparm[funcparmno] = value;
                funcparmno++;
                action = P_FuncRef1;
                return;
            case P_FuncCall:
                pos    = scopestack[scopelevel].retpos;
                value  = 0;
                action = P_ValueEnd;
                return;
        }
        action = scopestack[scopelevel].action;
        if (skip == Skip_For2) skip = 0;
    }
}

bool CParser::MakeByteCode (void) {
    unsigned bcpos, i, n,
             value, theline = 1;
    char     word     [33],
             *spcword [1000];
    memset (bytecode, 0, 0xffff);
    memset (spcword, 0, sizeof (spcword));
    pos = 0;
    for (i=Str_W_Last + 1, n=0; i<Str_SysLast; i++, n++) if (StrTable[i]) {
        spcword[n] = strdup(StrTable[i]);
    }
    for (bcpos = 0; pos < scrlen; bcpos ++) {
        while (script[pos] <= ' ' || script[pos] > '~' || script[pos] == ';' || script[pos] == ':') {
            if (script[pos] == 0x0a) theline ++;
            pos++;
            if (pos >= scrlen) break;
        }
        if (pos >= scrlen) break;
        
        if (script[pos] >= '0' && script[pos] <= '9') {
            value = 0;
            while (script[pos] >= '0' && script[pos] <= '9') {
                value = (value * 10) + script[pos] - '0';
                pos++;
                if (pos >= scrlen) break;
            }
            linenum [bcpos]   = theline;
            bytecode[bcpos++] = Str_W_Const;
            linenum [bcpos]   = theline;
            bytecode[bcpos]   = value;
            
        } else if ((script[pos] >= 'a' && script[pos] <= 'z') || script[pos] == '_') {
            word[0] = script[pos];
            for (n=1; 1; n++) {
                if (n >= MAX_IDENT) {
                    PrintError (Str_S_LongID);
                    return false;
                }
                if ((script[pos + n] < 'a' || script[pos + n] > 'z')
                &&  (script[pos + n] < '0' || script[pos + n] > '9')
                &&   script[pos + n] != '_') break;
                if (pos + n >= scrlen) break;
                word[n] = script[pos + n];
            }
            word[n] = 0;
            linenum [bcpos] = theline;
            bytecode[bcpos] = P_Unknown;
            pos += n;
            for (i=Str_W_First+1; i<Str_W_Const; i++) {
                if (strcmp (StrTable[i], word) == 0) {
                    bytecode[bcpos] = i;
                    break;
                }
            }
            if (bytecode[bcpos] == P_Unknown) for (i=Str_W_Const+1; i<Str_W_Nop; i++) {
                if (strcmp (StrTable[i], word) == 0) {
                    bytecode[bcpos] = Str_W_Const;
                    value = Constant (i);
                    break;
                }
            }
            if (bytecode[bcpos] == P_Unknown) {
                for (i=0; i<1000; i++) {
                    if (spcword[i] == NULL) {
                        spcword[i] = strdup (word);
                        bytecode[bcpos] = Str_W_Last + i + 1;
                        break;
                    }
                    if (strcmp (spcword[i], word) == 0) {
                        bytecode[bcpos] = Str_W_Last + i + 1;
                        break;
                    }
                }
            }
            
        } else {
            for (i=Str_W_Nop; i<Str_W_Last; i++) {
                if (script[pos] == StrTable[i][0]) {
                    if (StrTable[i][1] == script[pos + 1]) {
                        if (StrTable[i][2] == script[pos + 2]) pos++;
                        else if (StrTable[i][2] > 32) continue;
                        pos++;
                    } else if (StrTable[i][1] > 32) continue;
                    pos++;
                    linenum [bcpos] = theline;
                    bytecode[bcpos] = i;
                    break;
                }
            }
        }
        
        if (bytecode[bcpos] == Str_W_Comment
        ||  bytecode[bcpos] == Str_W_Comm1) {
            while (pos < scrlen && script[pos] >= ' ') pos++;
            bcpos --;
        }
        if (bytecode[bcpos] == Str_W_ComLong) {
            while (pos < scrlen - 1 && (script[pos] != '*' || script[pos + 1] != '/')) {
                if (script[pos] == 0x0a) theline ++;
                pos++;
            }
            pos += 2;
            bcpos --;
        }
        if (bytecode[bcpos] == Str_W_Quote) {
            while (pos < scrlen - 1 && script[pos] != '"')
                bytecode[++bcpos] = script[++pos];
            bytecode[bcpos] = '"';
            pos   ++;
            bcpos ++;
        }
        
        if (tracelevel >= 4 || 1) {
            if (bytecode[bcpos - 1] == Str_W_Const) {
                TRACE3 ("%3.3d %4.4x '%d'\n", theline, 0, bytecode[bcpos]);
            } else if (bytecode[bcpos] > Str_W_Last) {
                TRACE3 ("%3.3d %4.4x '%s'\n", theline, bytecode[bcpos],
                        spcword[bytecode[bcpos] - Str_W_Last - 1]);
            } else {
                    TRACE3 ("%3.3d %4.4x %s\n", theline, bytecode[bcpos], StrTable[bytecode[bcpos]]);
            }
        }
    }
    if (tracelevel >= 2) {
        for (i=0; i<1000; i++) if (spcword[i])
            TRACE2 ("%3.3d %s\n", i + 1, spcword[i]);
    }
    bytecode[bcpos] = Str_W_Last;
    linenum [bcpos] = theline;
    pos = 0;
    return true;
}

unsigned CParser::Process (void) {
    unsigned token = 0, // newword = 0,
             i,
             n;
    char     word  [33],
             pbuf  [255];
    CVar     *temp;
    if (error) return error;
    if (end)   return Str_S_None;

    // cut here
    if (pos >= scrlen) {
        token = Str_W_Last;
    } else {
        while (script[pos] <= ' ' || script[pos] > '~' || script[pos] == ';' || script[pos] == ':') {
            pos++;
            if (pos >= scrlen) {
                token = Str_W_Last;
                break;
            }
        }
    }
    if (token == 0) {
        if (script[pos] >= '0' && script[pos] <= '9') {
            value = 0;
            while (script[pos] >= '0' && script[pos] <= '9') {
                if (action != P_Switch)
                    value = (value * 10) + script[pos] - '0';
                pos++;
                if (pos >= scrlen) {
                    token = Str_W_Last;
                    break;
                }
            }
            if (value > 0xffff) return Str_S_Ovflow;
            if (token == 0) token = Str_W_Const;
        } else if ((script[pos] >= 'a' && script[pos] <= 'z') || script[pos] == '_') {
            word[0] = script[pos];
            for (n=1; 1; n++) {
                if (n >= MAX_IDENT) return Str_S_LongID;
                if ((script[pos + n] < 'a' || script[pos + n] > 'z')
                &&  (script[pos + n] < '0' || script[pos + n] > '9')
                &&   script[pos + n] != '_') break;
                if (pos + n >= scrlen) break;
                word[n] = script[pos + n];
            }
            word[n] = 0;
            token = P_Unknown;
            pos += n;
            for (i=Str_W_First+1; i<Str_W_Const; i++) {
                if (strcmp (StrTable[i], word) == 0) {
                    token = i;
                    break;
                }
            }
            if (token == P_Unknown) for (i=Str_W_Const+1; i<Str_W_Nop; i++) {
                if (strcmp (StrTable[i], word) == 0) {
                    token = Str_W_Const;
                    value = Constant (i);
                    break;
                }
            }
            if (token == P_Unknown) for (temp = varlist; temp != NULL; temp = temp->next) {
                if (strcmp (temp->name, word) == 0) {
                    if (temp->is_func) token = P_FuncRef;
                    else               token = P_VarRef;
                    varptr = &temp->value;
                    break;
                }
            }
            if (token == P_Unknown) {
                i = GetSpcToken (word);
                if (i) {
                    token = Str_W_Const;
                    value = i;
                }
            }
        } else {
            for (i=Str_W_Nop; i<Str_W_Last; i++) {
                if (script[pos] == StrTable[i][0]) {
                    if (StrTable[i][1] == script[pos + 1]) {
                        if (StrTable[i][2] == script[pos + 2]) pos++;
                        else if (StrTable[i][2] > 32) continue;
                        pos++;
                    } else if (StrTable[i][1] > 32) continue;
                    pos++;
                    token = i;
                    break;
                }
            }
        }
    }
    // to here
    /*
    if (pos >= scrlen) {
        token = Str_W_Last;
    } else {
        token             = bytecode[pos];
        curline = linenum [pos];
        pos ++;
        if (token == Str_W_Const) {
            value = bytecode[pos];
            pos ++;
        } else if (token > Str_W_Last) {
            newword = token - Str_W_Last;
            token = P_Unknown;
            for (temp = varlist; temp != NULL; temp = temp->next) {
                if (temp->tokenref == newword) {
                    if (temp->is_func) token = P_FuncRef;
                    else               token = P_VarRef;
                    varptr = &temp->value;
                    break;
                }
            }
        }
    }
    */

    switch (token) {
        case Str_W_Comm1:
        case Str_W_Comment: while (pos < scrlen && script[pos] >= 32) pos++;
                            return 0;
        case Str_W_ComLong: while (pos < scrlen - 1 && (script[pos] != '*' || script[pos + 1] != '/'))
                                pos++;
                            pos += 2;
                            return 0;
        case Str_W_Quote:   for (i=0; i<sizeof (pbuf) - 1; i++) {
                                if (script[pos] == '\\' && script[pos + 1] == 'n') {
                                    pbuf[i] = 0x0a;
                                    pos += 2;
                                } else {
                                    pbuf[i] = script[pos];
                                    pos ++;
                                }
                                if (pos >= scrlen - 1 || script[pos] < 32 || script[pos] == '"') break;
                            }
                            pbuf[i + 1] = 0;
                            if (i == sizeof (pbuf) - 1) {
                                while (pos < scrlen - 1 && script[pos] != '"' && script[pos] >= ' ')
                                    pos ++;
                            }
                            pos ++;
                            break;
        case Str_W_InEq1:
        case Str_W_InEq2:
        case Str_W_InEq3:   token = Str_W_InEqual; break;
        case Str_W_GrEq1:   token = Str_W_GreatEq; break;
        case Str_W_LsEq1:   token = Str_W_LessEq;  break;
        case Str_W_Equals1: token = Str_W_Equals;  break;
        case Str_W_Func1:   token = Str_W_Func;    break;
        case Str_W_Div1:    token = Str_W_Divide;  break;
        case Str_W_Mod1:    token = Str_W_Modulo;  break;
        case Str_W_And1:    token = Str_W_LogAnd;  break;
        case Str_W_Or1:     token = Str_W_LogOr;   break;
        case Str_W_Not1:    token = Str_W_Not;     break;
        case Str_W_Xor1:    token = Str_W_Xor;     break;
        case Str_W_Xor2:    token = Str_W_Xor;     break;
        case Str_W_Var1:    token = Str_W_Var;     break;
        case Str_W_Begin:   token = Str_W_BrackL;  break;
        case Str_W_End:     token = Str_W_BrackR;  break;
        case Str_W_Print1:  token = Str_W_Print;   break;
        case 0:             return Str_S_Unknown;
    }
    if (tracelevel >= 3) switch (token) {
        case P_Unknown: Trace (3, StrTable[Str_S_TNew],  word); break;
        case P_FuncRef: Trace (3, StrTable[Str_S_TFunc], word); break;
        case P_VarRef:  Trace (3, StrTable[Str_S_TVar],  word); break;
        default:        Trace (3, StrTable[Str_S_Token], StrTable[token]);
    }
    temp = (CVar*) varptr;
    while (theApp.runlevel < Run_Abort) switch (action) {
        case Str_W_Main:
            end = 1;
            if (token != Str_W_Last) return Str_S_NoEOF;
            return Str_S_None;
        case P_Start:
            switch (token) {
                case Str_W_Var:  PushScope (P_NewVar);   return 0;
                case Str_W_Func: PushScope (P_NewFunc);  return 0;
                case Str_W_Main: action = Str_W_Main;
                                 if (mainpos == 0) skip = Skip_Stack;
                                 mainpos = pos;
                                 Trace (1, StrTable[Str_S_Main]);
                                 PushScope (P_Command);  return 0;
                case Str_W_Last: return Str_S_NoMain;
            }
            return Str_S_Start;
        case P_CmdVar:
            switch (token) {
                case P_VarRef:
                case P_FuncRef:  action = token; return 0;
            }
            return Str_S_VarCmd;
        case P_Command:
            switch (token) {
                case Str_W_Var:                       PushScope (P_NewVar);   return 0;
                case Str_W_Switch: return Str_S_SwNest; // disables switch
                    if (scopestack[scopelevel - 1].action == P_CaseRun) return Str_S_SwNest;
                case Str_W_While:  retpos = pos;
                case Str_W_Case:
                case Str_W_Return:
                case Str_W_Print:
                case Str_W_If:     action = token;    PushScope (P_Value);    return 0;
                case Str_W_Do:     retpos = pos;
                case Str_W_BrackL: action = token;    PushScope (P_Command);  return 0;
                case P_VarRef:
                case P_FuncRef:
                case Str_W_For:    action = token;                            return 0;
                case Str_W_Default:                                           return 0;
                case Str_W_Func:   return Str_S_Func;
                case Str_W_Quote:  if (skip == 0) Output (pbuf); return 0;
                case Str_W_Endl:   if (skip == 0) Output ("\n"); return 0;
                case Str_W_Cont:
                case Str_W_Break:
                    for (i=scopelevel-1; i>0; i--) switch (scopestack[i].action) {
                        case P_CaseRun:
                            if (token == Str_W_Cont) continue;
                            if (skip == 0) {
                                scopelevel = i - 1;
                                pos        = scopestack[i - 1].retpos;
                                action     = Str_W_Switch;
                                skip       = Skip_Break;
                                PushScope (P_Value);
                            }
                            return 0;
                        case P_Loop:
                            if (skip == 0) {
                                scopelevel = i;
                                pos        = scopestack[i].retpos;
                                action     = Str_W_While;
                                if (token == Str_W_Break) skip = Skip_Break;
                                PushScope (P_Value);
                            }
                            return 0;
                        case P_For3:
                            if (skip == 0) {
                                scopelevel = i;
                                pos = scopestack[scopelevel].retpos;
                                action = P_For4;
                                if (token == Str_W_Break) skip = Skip_Break;
                                PushScope (P_Command);
                                skip = Skip_For2;
                                PushScope (P_Value);
                                if (token == Str_W_Break) skip = Skip_Break;
                                PushScope (P_Command);
                            }
                            return 0;
                    }
                    if (token == Str_W_Cont) return Str_S_Cont;
                    return Str_S_Break;
                case Str_W_BrackR:
                    if (scopestack[scopelevel - 1].action != Str_W_BrackL
                    &&  scopestack[scopelevel - 1].action != P_CaseRun)   return Str_S_RgtBrac;
                    scopelevel--;
                    PopScope ();
                    KillVars ();
                    return 0;
                case Str_W_Last: return Str_S_EOF;
            }
            return Str_S_Cmd;
        case P_SwBreak_0:
        case P_Switch_0:
            if (token != Str_W_BrackL) return Str_S_SwBrac;
            action++;
            return 0;
        case P_SwBreak:
            if (token == Str_W_BrackR) PopScope ();
            if (token == Str_W_BrackL) {
                PushScope (Str_W_BrackL);
                PushScope (P_Command);
            }
            return 0;
        case P_Switch:
            switch (token) {
                case Str_W_BrackL:  PushScope (Str_W_BrackL);
                                    skip = Skip_While;
                                    PushScope (P_Command);
                                    return 0;
                case Str_W_Var:     PushScope (P_NewVar);
                                    return 0;
                case Str_W_Case:    action = P_Case;
                                    PushScope (P_Value);
                                    return 0;
                case Str_W_BrackR:  PopScope ();
                                    return 0;
                case Str_W_Switch:  return Str_S_SwNest;
                case Str_W_Default: action = P_CaseRun;
                                    PushScope (P_Command);
                                    return 0;
            }
            return 0;
        case Str_W_For:
            if (token != Str_W_ParenL) return Str_S_ForParn;
            action = P_For0;
            retpos = pos;
            PushScope (P_Command);
            return 0;
        case P_For2:
            if (token == Str_W_ParenR) {
                action = P_For3;
                PushScope (P_Command);
                return 0;
            }
            return Str_S_ForParn;
        case P_For3:
            if (skip == 0) {
                pos = scopestack[scopelevel].retpos;
                action = P_For4;
                PushScope (P_Command);
                skip = Skip_For2;
                PushScope (P_Value);
                PushScope (P_Command);
                return 0;
            }
            if (skip == Skip_For) skip = 0;
            PopScope ();
            break;
        case Str_W_Do:
            if (token != Str_W_While) return Str_S_While;
            action = P_Do;
            PushScope (P_Value);
            return 0;
        case P_Do:
            if (skip == 0 && value != 0) {
                pos = retpos;
                action = Str_W_Do;
                PushScope (P_Command);
                return 0;
            }
            PopScope ();
            break;
        case P_Loop:
            if (skip == 0) {
                pos = scopestack[scopelevel].retpos;
                action = Str_W_While;
                PushScope (P_Value);
                return 0;
            }
            if (skip == Skip_While) skip = 0;
            PopScope ();
            break;
        case P_Then:
            if (token == Str_W_Else) {
                action = Str_W_Else;
                if      (skip == Skip_If) skip = 0;
                else if (skip == 0)       skip = Skip_If;
                PushScope (P_Command);
                return 0;
            }
            PopScope ();
            break;
        case P_VarRef:
            if (temp->arraysize) {
                if (token != Str_W_ArrayL) return Str_S_Array;
                PushScope (Str_W_ArrayL);
                PushScope (P_Value);
                return 0;
            }
            if (token >= Str_W_Assign && token <= Str_W_Dec) {
                action = token;
                PushScope (P_Value);
                if (token == Str_W_Inc || token == Str_W_Dec) PopScope ();
                return 0;
            }
            return Str_S_ReqAsg;
        case P_Value:
            switch (token) {
                case P_VarRef:     if (temp->arraysize == 0) {
                                       value = *varptr; break;
                                   }
                                   action = P_ArrayRef;
                                   return 0;
                case P_FuncRef:    action = token; return 0;
                case Str_W_Const:  break;
                case Str_W_ParenL:
                case Str_W_Not:    action = token;        PushScope (P_Value); return 0;
                case Str_W_Minus:  action = Str_W_Negate; PushScope (P_Value); return 0;
                default:           return Str_S_Expr;
            }
            action = P_ValueEnd;
            return 0;
        case P_ValueEnd:
            if (token >= Str_W_LogAnd && token <= Str_W_Or) {
                i = scopestack[scopelevel - 1].action;
                if (i >= Str_W_LogAnd && i <= Str_W_Or && i > token) PopScope ();
                action = token;
                PushScope (P_Value);
                return 0;
            } else {
                PopScope ();
                if (token == Str_W_ParenR) {
                    while (action == P_ValueEnd) PopScope ();
                    if (action == P_FuncRef1) break;
                    if (skip == Skip_For1) return 0;
                    if (action == P_For2) {
                        action = P_For3;
                        PushScope (P_Command);
                    } else {
                        if (action != Str_W_ParenL) return Str_S_RgtParn;
                        action = P_ValueEnd;
                    }
                    return 0;
                } else {
                    if (action == P_For2) return Str_S_ForParn;
                    if (action == Str_W_ParenL) return Str_S_RgtParn;
                }
                break;
            }
        case Str_W_Print:
            if (value & 0x80000000) value = 0;
            if (value & 0x00010000) value = 0xffff;
            if (skip == 0) Output (MkStr (value));
            action = P_Command;
            break;
        case P_NewVar:
            if (token != P_Unknown) {
                if (temp == NULL)                  return Str_S_DupDef;
                if (temp->is_func)                 return Str_S_DupDef;
                if (temp->scope >= scopelevel - 1) return Str_S_DupDef;
            }
            AddVar (word, P_NewVar); // AddVar (newword, P_NewVar);
            action++;
            return 0;
        case P_NewVarEnd:
            switch (token) {
                case Str_W_Comma:  action = P_NewVar; return 0;
                case Str_W_Assign: PushScope (Str_W_Assign);
                                   varptr = scopestack[scopelevel - 1].varptr;
                                   PushScope (P_Value); return 0;
                case Str_W_ArrayL: action = P_NewArrayEnd;
                                   PushScope (P_Value); return 0;
            }
            PopScope ();
            break;
        case P_NewArrayEnd:
            if (token != Str_W_ArrayR) return Str_S_RgtBrck;
            if (value < 2 || value > 4096) return Str_S_ArBound;
            temp->arraysize = value;
            temp->arraydata = new unsigned [value];
            if (temp->arraydata == NULL) {
                theApp.runlevel = Run_MemError;
                return 0;
            }
            memset (temp->arraydata, 0, value * sizeof (unsigned));
            action = P_NewArrayEnd1;
            return 0;
        case P_NewArrayEnd1:
            if (token == Str_W_Comma) {
                action = P_NewVar; return 0;
            }
            PopScope ();
            break;
        case Str_W_ArrayL:
            if (token != Str_W_ArrayR) return Str_S_RgtBrck;
            if (skip == 0) {
                temp = (CVar*) scopestack[scopelevel].varptr;
                if (value >= temp->arraysize) return Str_S_ArrayOv;
                varptr = temp->arraydata + value;
            }
            PopScope ();
            action = P_ArrayAsg;
            return 0;
        case P_ArrayAsg:
            if (token >= Str_W_Assign && token <= Str_W_Dec) {
                action = token;
                PushScope (P_Value);
                if (token == Str_W_Inc || token == Str_W_Dec) PopScope ();
                return 0;
            }
            return Str_S_ReqAsg;
        case P_ArrayRef:
            if (token != Str_W_ArrayL) return Str_S_Array;
            PushScope (P_ArrayRef1);
            PushScope (P_Value);
            return 0;
        case P_ArrayRef1:
            if (token != Str_W_ArrayR)    return Str_S_RgtBrck;
            if (skip == 0) {
                temp = (CVar*) scopestack[scopelevel].varptr;
                if (value >= temp->arraysize) return Str_S_ArrayOv;
                value = temp->arraydata [value];
            }
            PopScope ();
            action = P_ValueEnd;
            return 0;
        case P_NewFunc:
            if (token != P_Unknown) return Str_S_DupDef;
            AddVar (word, P_NewFunc); // newword, 
            action = P_NewFuncEnd;
            return 0;
        case P_NewFuncEnd:
            switch (token) {
                case Str_W_ParenL: action = P_FuncParm; return 0;
                case Str_W_BrackL: action = Str_W_BrackL;
                                   skip   = Skip_Func;
                                   PushScope (P_Command);
                                   return 0;
            }
            return Str_S_FuncDef;
        case P_FuncParm:
            switch (token) {
                case Str_W_Comma:   return 0;
                case Str_W_ParenR:  action = P_NewFuncEnd; return 0;
                case P_Unknown:     temp->value++;
                                    if (temp->value >= MAX_PARM) return Str_S_ParmOv;
                                    scopelevel++;
                                    AddVar (word, P_NewVar); // newword, 
                                    scopelevel--;
                                    varptr = (unsigned*) temp;
                                    return 0;
            }
            return Str_S_DupDef;
        case P_FuncRef:
            if (token != Str_W_ParenL) return Str_S_FuncCl;
            action     = P_FuncRef1;
            funcptr    = temp;
            funcparmno = 0;
            return 0;
        case P_FuncRef1:
            switch (token) {
                case Str_W_Comma:   return 0;
                case Str_W_ParenR:  if (funcparmno > funcptr->value) return Str_S_FuncPrM;
                                    if (funcparmno < funcptr->value) return Str_S_FuncPrF;
                                    if (skip) {
                                        value = 0;
                                        action = P_ValueEnd;
                                        return 0;
                                    }
                                    if (funcptr->arraysize) {
                                        value  = SysCall (funcptr->arraysize);
                                        action = P_ValueEnd;
                                        return 0;
                                    }
                                    action = P_FuncEntry;
                                    retpos = pos;
                                    value  = 0;
                                    pos    = funcptr->scope;
                                    return 0;
                case P_FuncRef:     return Str_S_FuncNst;
            }
            PushScope (P_Value);
            value = scopestack[scopelevel - 1].value;
            break;
        case P_FuncEntry:
            if (token != Str_W_ParenL) return Str_S_FuncDef;
            action = P_FuncEnt1;
            return 0;
        case P_FuncEnt1:
            switch (token) {
                case P_VarRef:
                case P_Unknown:    scopelevel += 2;
                                   AddVar (word, P_NewVar); // newword
                                   ((CVar*) varptr)->value = funcparm[value];
                                   value++;
                                   scopelevel -= 2;
                                   return 0;
                case Str_W_Comma:  return 0;
                case Str_W_ParenR: action = P_FuncCall;
                                   PushScope (P_Command);
                                   return 0;
                case P_FuncRef:    return Str_S_VarFunc;
            }
            return Str_S_FuncDef;
        case Str_W_Return:
            if (skip == 0) {
                for (i=scopelevel-1; i>0; i--) {
                    if (scopestack[i].action == P_FuncCall) {
                        scopelevel = i;
                        pos = scopestack[i].retpos;
                        KillVars ();
                        action = P_ValueEnd;
                        return 0;
                    }
                }
                end = 1;
                return 0;
            }
            action = P_ValueEnd;
            break;
        default:
            if (action <= P_Unknown) {
                StrTable[Str_S_Fail] = MkStr (action);
                return Str_S_Fail;
            }
            return action;
    }
    return Str_ErrLast;
}


#endif
